/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*! \file basecont.h
	\brief Container classes

	These classes provide simple containers, similar to what is provided
	in STL. The reason for not using STL here is mostly due to the lack
	of partial template specialization in some target compilers, thus 
	making serialization of STL containers impractical.
	For std::vector, there is a workaround base::vector in base.h, but it
	is not very clean (just required to ensure DPS will compile with
	compilers that do not support partial template specialization).
*/

#ifndef INCLUDED_DPS_CONTAINER_H
#define INCLUDED_DPS_CONTAINER_H

namespace base
{
	//! Fixed size array
	/*! \ingroup internal
		This class is used as a substitute for real array when using ARRAY
		declarations in universal serialization.
	*/
	template<typename el, int size> class Array
	{
	private:
		el items[size];

	public:
		operator el*() { return items; }

		//! Serialize pointed object
		void write(OutputStream *s) const
		{
			for(Size i=0;i<size;i++)
			{
				ENTERITEM("element");
				typehandler::write(s,items[i]);
				LEAVEITEM("element");
			}
		}
		//! Deserialize pointed object
		void read(InputStream *s)
		{
			for(Size i=0;i<size;i++)
			{
				ENTERITEM("element");
				typehandler::read(s,items[i]);
				LEAVEITEM("element");
			}
		}
		//! Get size of array
		static Size getSize() { return size; }
	};
	//! Variable size vector of references (object pointers)
	/*! \ingroup user
		This is a standard vector class, similar to Java's vector, or
		an STL std::vector containing smart pointers.
	*/
	template<typename el, bool ref = true> class Vector : public Object
	{
	protected:
		//! Items in the vector
		el **items;
		//! Number of items currently in vector
		Size cnt;
		//! Maximum capacity of vector
		Size maxCnt;
	public:
		//! Constructor for an empty vector
		Vector() 
		{ 
			cnt=0; 
			maxCnt=10; 
			items=(el**)malloc(maxCnt*sizeof(el *)); 
		}
		//! Copy constructor
		Vector(const Vector& src) : Object()
		{
			cnt=0;
			maxCnt=src.maxCnt;
			items=(el**)malloc(maxCnt*sizeof(el *)); 
			for(Size j=0;j<src.cnt;j++) 
				add(src.items[j]); 
		}
		//! Destructor
		~Vector()
		{
			if(items)
			{
				if(ref)
				{
					for(Size i=0;i<cnt;i++)
						items[i]->downcount();
				}
				free(items);
			}
		}
		
		//! Assignment operator
		Vector& operator=(const Vector& src)
		{
			clear();
			for(Int32 j=0;j<src.cnt;j++) 
				add(src.items[j]); 
			return *this;
		}

		//! Add item to vector
		Size add(el *i) 
		{ 
			if(cnt==maxCnt) 
			{ 
				maxCnt+=10; 
				items=(el **)realloc(items,maxCnt*sizeof(el *)); 
			} 
			if(ref)
				i->upcount();
			items[cnt++]=i; 
			return cnt-1;
		}

		//! Add item to vector
		Size add(Ptr<el>& i) 
		{ 
			if(cnt==maxCnt) 
			{ 
				maxCnt+=10; 
				items=(el **)realloc(items,maxCnt*sizeof(el *)); 
			} 
			if(ref)
				i->upcount();
			items[cnt++]=i; 
			return cnt-1;
		}

		//! Add item to vector at specified position
		Size add(el *i, int off) 
		{ 
			if(cnt==maxCnt) 
			{ 
				maxCnt+=10; 
				items=(el **)realloc(items,maxCnt*sizeof(el *)); 
			} 
			if(ref)
				i->upcount();
			memmove(&items[off+1],&items[off],(cnt-off)*sizeof(el*));
			items[off]=i; 
			cnt++;
			return off;
		}

		//! Add item to vector only if it is not already in vector
		Size adds(el *i) 
		{ 
			if(cnt==maxCnt) 
			{ 
				maxCnt+=10; 
				items=(el **)realloc(items,maxCnt*sizeof(el *)); 
			} 
			Size j;
			for(j=0;j<cnt;j++)
			{
				if(items[j]==i)
					break;
			}
			if(j>=cnt)
			{
				if(ref)
					i->upcount();
				items[cnt++]=i; 
				return cnt-1;
			}
			return -1;
		}

		//! Add item to vector only if it is not already in vector
		Size adds(Ptr<el>& i) 
		{ 
			if(cnt==maxCnt) 
			{ 
				maxCnt+=10; 
				items=(el **)realloc(items,maxCnt*sizeof(el *)); 
			} 
			Size j;
			for(j=0;j<cnt;j++)
			{
				if(items[j]==i)
					break;
			}
			if(j>=cnt)
			{
				if(ref)
					i->upcount();
				items[cnt++]=i; 
				return -1;
			}
			return j;
		}

		//! Add all items in parameter vector to this vector
		void add(Vector& i) 
		{ 
			for(Size j=0;j<i.cnt;j++) 
				add(i.items[j]); 
		}

		//! Add all items in parameter vector to this vector if not already in this vector
		void adds(Vector& i) 
		{ 
			for(Int32 j=0;j<i.cnt;j++) 
				adds(i.items[j]); 
		}

		//! Remove item from vector
		bool remove(el *i)
		{
			for(Size j=0;j<cnt;j++)
			{
				if(items[j]==i)
				{
					if(ref)
						i->downcount();
					memmove(&items[j],&items[j+1],(cnt-j-1)*sizeof(el *));
					cnt--;
					return true;
				}
			}
			return false;
		}

		//! Remove item at index from vector
		bool remove(Size j)
		{
			if(ref)
				items[j]->downcount();
			memmove(&items[j],&items[j+1],(cnt-j-1)*sizeof(el *));
			cnt--;
			return true;
		}

		//! Get item from vector
		el* operator[](Size i) { return items[i]; }

		//! Get size of vector
		Size size() const { return cnt; }

		//! Clear vector
		void clear() 
		{ 
			if(ref)
			{
				for(Size i=0;i<cnt;i++)
					items[i]->downcount();
			}
			cnt=0; 
		}

		//! Get index of item in vector
		Size indexOf(el *i) const 
		{
			for(Size j=0;j<cnt;j++)
			{
				if(items[j]==i)
				{
					return j;
				}
			}
			return -1;
		}
		
		//! Serialize vector
		void write(OutputStream *s) const
		{
			ENTERITEM("size");
			XPSize vs=(XPSize)size();
			typehandler::write(s,vs);
			LEAVEITEM("size");
			for(XPSize i=0;i<vs;i++)
			{
				ENTERITEM("element");
				s->writeRef(items[i]);
				LEAVEITEM("element");
			}
		}

		//! Deserialize vector
		void read(InputStream *s)
		{
			ENTERITEM("size");
			XPSize vs;
			typehandler::read(s,vs);
			LEAVEITEM("size");
			clear();
			for(XPSize i=0;i<vs;i++)
			{
				ENTERITEM("element");
				add((el*)s->readRef());
				LEAVEITEM("element");
			}
		}
	};

	//! Memory block of a simple type
	/*! \ingroup user
		This class provides a wrapper around memory blocks containing items of
		a specific type. For instance, the following can be replaced:
		<br><pre>Int32 *data=new Int32[itemCount];</pre>
		<br>by
		<br><pre>Buffer<Int32> data(itemCount);</pre>
		<br>The buffer overloads standard cast operators, so its use should be
		fairly transparent when a traditional allocation is replaced.
		<p>This class can be serialized.
	*/
	template<typename bt> class Buffer : public Object
	{
		//! Pointer to data
		void *data;
		//! Number of items in memory block
		Size len;
	public:
		//! Constructor for a zero-length memory block
		Buffer() { data=NULL; len=0; }

		//! Constructor for a memory block
		/*! \param l Initial size of memory block
		*/
		Buffer(Size l) { data=malloc(l*sizeof(bt)); len=l; }

		//! Constructor for a memory block
		/*! \param src Pointer to data to copy to memory block
			\param l   Size of memory block pointed by src
		*/
		Buffer(const bt *src, Size l) 
		{ 
			data=NULL; 
			len=0; 
			if(src && l) 
			{ 
				data=malloc(l*sizeof(bt)); 
				memcpy(data,src,l*sizeof(bt)); 
				len=l; 
			} 
		}


		//! Copy constructor for a memory block
		/*! \param src Memory block to copy data from
		*/
		Buffer(const Buffer &src) 
		{ 
			data=NULL; 
			len=0; 
			if(src.data) 
			{ 
				data=malloc(src.len*sizeof(bt)); 
				memcpy(data,src.data,src.len*sizeof(bt)); 
				len=src.len; 
			} 
		}

		//! Destructor
		~Buffer() { if(data) free(data); }

		//! Get size of memory block
		Size size() { return len; }
		
		//! Copy a memory block into this one
		/*! \param src Source memory block
		*/
		Buffer& operator=(const Buffer& src) 
		{ 
			if(data) 
				free(data); 
			data=NULL; 
			len=0; 
			if(src.data) 
			{ 
				data=malloc(src.len*sizeof(bt)); 
				check(data!=NULL,("Could not allocate %d bytes",src.len*sizeof(bt)));
				memcpy(data,src.data,src.len*sizeof(bt)); 
				len=src.len; 
			} 
			return *this; 
		}

		//! Cast to data type pointer
		operator bt *() { return (bt*)data; }

		//! Get pointer to data
		bt *get() { return (bt*)data; }

		//	bt& operator[](Int32 i) { return ((bt*)data)[i]; }
		//! Resize memory block
		/*! \param newSize New size of memory block
		*/
		void resize(Size newSize) 
		{ 
			if(newSize) 
			{ 
				if(data) 
					data=realloc(data,newSize*sizeof(bt)); 
				else 
					data=malloc(newSize*sizeof(bt)); 
				len=newSize; 
				check(data!=NULL,("Could not allocate %d bytes",newSize*sizeof(bt)));
			} 
			else 
			{ 
				if(data) 
					free(data); 
				len=0; 
				data=NULL; 
			} 
		}
		
		//! Serialize memory block
		void write(OutputStream *s) const
		{
			ENTERITEM("size");
			XPSize tl=(XPSize)len;
			typehandler::write(s,tl);
			LEAVEITEM("size");
			ENTERITEM("data");
			s->writeArray(data,len,sizeof(bt));
			LEAVEITEM("data");
		}

		//! Deserialize memory block
		void read(InputStream *s)
		{
			if(data)
				free(data);
			data=NULL;
			len=0;
			ENTERITEM("size");
			XPSize rlen;
			typehandler::read(s,rlen);
			LEAVEITEM("size");
			len=rlen;
			if(len) 
			{ 
				ENTERITEM("data");
				data=malloc(len*sizeof(bt)); 
				s->readArray(data,len,sizeof(bt));
				LEAVEITEM("data");
			} 
		}
		
		//! Identification
		CUSTOMSERIAL(Buffer);
	};

	//! String pair
	/*! \ingroup user
	*/
	class StringPair : public Object
	{
		CLASSDEF(StringPair)
			BASECLASS(base::Object)
		MEMBERS
			//! Name string
			ITEM(std::string,name)
			//! Value string
			ITEM(std::string,value)
		CLASSEND;

		//! Constructor
		StringPair(const char *n = NULL, const char *v = NULL) 
		{ 
			if(n) name=n; else name="(null)";
			if(v) value=v; else value="(null)";
		}
	};


	//! String map (vector of string pairs)
	/*! \ingroup user
	*/
	class StringMap : public Object
	{
		//! Vector of string pairs in map
		Vector<StringPair> pairs;
	public:
		//! Constructor
		StringMap() { }

		//! Constructor from command line args
		StringMap(Int32 argc, char *argv[]);

		//! Constructor from file contents
		StringMap(const char *fname);

		//! Constructor from vector of STL strings
		StringMap(std::vector<std::string> args);

		//! Add pair to map
		void add(StringPair *p) { pairs.add(p); }

		//! Find given name in map
		bool isSet(const char *n) { for(Size i=0;i<pairs.size();i++) if(pairs[i]->name==n) return true; return false; }

		//! Get value for given name in map
		/*! \param name Name to search
			\param defVal Default value if name is not found
		*/
		const char *getValue(const char *name, const char *defVal=NULL) { for(Size i=0;i<pairs.size();i++) if(pairs[i]->name==name) return pairs[i]->value.c_str(); return defVal; }

		//! Return size of map
		Size size() { return pairs.size(); }

		//! Get indexed pair
		StringPair *getPair(Size i) { return pairs[i]; }
		
		IDENTIFYNOREG(StringMap);
	};

	//! FIFO queue of objects
	/*! \ingroup internal
		Items added to the queue must contain a member named next of type el*
	*/
	template<typename el> class Queue : public Object
	{
		IDENTIFYNOREG(Queue);

		//! First item in queue
		el *head;
		//! Last item in queue
		el *tail;
	public:
		//! Constructor
		Queue() 
		{ 
			head=tail=NULL; 
		}

		//! Push item at tail of queue
		void push(el *e) 
		{
			e->upcount(); 
			if(tail) 
			{ 
				tail->next=e; 
				e->next=NULL; 
				tail=e; 
			} 
			else 
			{ 
				head=tail=e; 
				e->next=NULL; 
			} 
		}

		//! Push item at tail of queue
		void push(Ptr<el>& e) 
		{
			e->upcount(); 
			if(tail) 
			{ 
				tail->next=e; 
				e->next=NULL; 
				tail=e; 
			} 
			else 
			{ 
				head=tail=e; 
				e->next=NULL; 
			} 
		}

		//! Push item at head of queue
		void pushHead(el *e) 
		{
			e->upcount(); 
			if(head==NULL) 
				tail=e;
			e->next=head;
			head=e; 
		}

		//! Push item at head of queue
		void pushHead(Ptr<el>& e) 
		{
			e->upcount(); 
			if(head==NULL) 
				tail=e;
			e->next=head;
			head=e; 
		}

		//! Pop item at head of queue
		Ptr<el> pop() 
		{ 
			if(head) 
			{ 
				Ptr<el> t=head; 
				head=head->next; 
				if(head==NULL) 
					tail=NULL; 
				t->downcount(); 
				return t; 
			} 
			return NULL; 
		}

		//! Get item at head of queue without removing it
		el* peek() { return head; }

		//! Get size of queue
		Size size() { Size cnt=0; for(el *t=head;t;t=t->next,cnt++); return cnt; }

		//! Get indexed item from queue
		el* operator[](Size i) { el *t=head; for(;t && i>0;t=t->next,i--); return t; }
	};

	//! Wrapper class for items added to SyncQueue
	/*! \ingroup internal
	*/
	template<typename el> class ItemHolder
	{
	public:
		//! Next item in queue
		ItemHolder *next;
		//! Pointed item
		Ptr<el> item;

		//! Constructor
		ItemHolder(el *it) { item=it; next=NULL; }
	};

	//! Queue with mutually exclusive access
	/*! \ingroup internal
	*/
	template<typename el> class SyncQueue : public Object
	{
		IDENTIFYNOREG(SyncQueue);
	
		//! First item
		ItemHolder<el> *head;

		//! Last item
		ItemHolder<el> *tail;

		//! Critical section for queue access
		CriticalSection critSec;
	public:
		//! Constructor
		SyncQueue() 
		{
			base::ScopedLock lock(critSec);
			head=tail=NULL; 
		}

		//! Add item to tail
		void push(el *it) 
		{
			base::ScopedLock lock(critSec);
			ItemHolder<el> *e=new ItemHolder<el>(it);
			if(tail) 
			{ 
				tail->next=e; 
				e->next=NULL; 
				tail=e; 
			} 
			else 
			{ 
				head=tail=e; 
				e->next=NULL; 
			} 
		}

		//! Add item to tail
		void push(Ptr<el>& it) 
		{
			base::ScopedLock lock(critSec);
			ItemHolder<el> *e=new ItemHolder<el>(it);
			if(tail) 
			{ 
				tail->next=e; 
				e->next=NULL; 
				tail=e; 
			} 
			else 
			{ 
				head=tail=e; 
				e->next=NULL; 
			} 
		}

		//! Get item from head
		Ptr<el> pop() 
		{ 
			base::ScopedLock lock(critSec);
			if(head) 
			{ 
				ItemHolder<el> *it=head;
				Ptr<el> t=it->item; 
				head=head->next; 
				if(head==NULL) 
					tail=NULL; 
				delete it;
				return t; 
			} 
			return NULL; 
		}

		//! Get item from head without removing it
		el* peek() { base::ScopedLock lock(critSec); return head->item; }

		//! Get size of queue
		Size size() { base::ScopedLock lock(critSec); Size cnt=0; for(el *t=head;t;t=t->next,cnt++); return cnt; }

		//! Is queue empty?
		Bool empty() { base::ScopedLock lock(critSec); return head==NULL; }
	};

//! A vector where all new objects are inserted at the end, and indexes stay constant (i.e. deleted items leave holes)
/*! \ingroup internal
	Items are always added to the end of this vector type. When items are 
	removed, it leaves a hole in the vector. This container class is
	specifically intended for collecting items that need a unique identifier
	(in this case the index in the container).
	This class is synchronized
*/
template<typename el> class SparseVector : public Object
{
	IDENTIFYNOREG(SparseVector);

	//! Items in vector
	el **items;
	//! Maximum size of vector
	Size maxSize;
	//! Current size of vector (including holes)
	Size cnt;
	//! Offset of first item
	Size off;
	//! Resizing increment
	Size incr;
	//! Critical section for synchronization
	CriticalSection critSec;
public:
	//! Constructor
	SparseVector(Size size0=10, Size incr0=10)
	{
		base::ScopedLock lock(critSec);

		maxSize=size0;
		off=0;
		cnt=0;
		incr=incr0;
		items=(el **)malloc(maxSize*sizeof(el*));
		memset(items,0,maxSize*sizeof(el*));
	}
	//! Destructor
	~SparseVector()
	{
		base::ScopedLock lock(critSec);

		if(items)
		{
			for(Size i=0;i<cnt;i++)
			{
				if(items[i]!=NULL)
					items[i]->downcount();
			}
			free(items);
		}
	}

	//! Add item to vector
	Size add(el *i)
	{
		base::ScopedLock lock(critSec);

		if(cnt>=maxSize)
		{
			maxSize+=incr;
			items=(el**)realloc(items,maxSize*sizeof(el*));
			memset(&items[maxSize-incr],0,incr*sizeof(el*));
		}
		if(i!=NULL)
			i->upcount();
		items[cnt++]=i;
		return off+cnt-1;
	}

	//! Add item to vector
	Size add(Ptr<el>& i)
	{
		base::ScopedLock lock(critSec);

		if(cnt>=maxSize)
		{
			maxSize+=incr;
			items=(el**)realloc(items,maxSize*sizeof(el*));
			memset(&items[maxSize-incr],0,incr*sizeof(el*));
		}
		if(i!=NULL)
			i->upcount();
		items[cnt++]=i;
		return off+cnt-1;
	}

	//! Add item to vectorat specified index
	Size add(el *i, Size ndx)
	{
		base::ScopedLock lock(critSec);

		if(ndx<off)
		{
			// Going backwards
			dump(0,"Backwards use of SparseVector not implemented yet!");
			return -1;
		}
		if(ndx-off>=maxSize)
		{
			Int32 pms=maxSize;
			maxSize=(maxSize+incr>ndx-off) ? (maxSize+incr) : (ndx+1-off);
			items=(el**)realloc(items,maxSize*sizeof(el*));
			memset(&items[pms],0,(maxSize-pms)*sizeof(el*));
		}
		if(ndx-off<cnt && items[ndx-off])
			items[ndx-off]->downcount();
		if(i!=NULL)
			i->upcount();
		items[ndx-off]=i;
		if(ndx-off+1>cnt)
			cnt=ndx-off+1;

		return ndx;
	}

	//! Add item to vector at specified index
	Size add(Ptr<el>& i, Int32 ndx)
	{
		base::ScopedLock lock(critSec);

		if(ndx<off)
		{
			// Going backwards
			dump(0,"Backwards use of SparseVector not implemented yet!");
			return -1;
		}
		if(ndx-off>=maxSize)
		{
			Int32 pms=maxSize;
			maxSize=(maxSize+incr>ndx-off) ? (maxSize+incr) : (ndx+1-off);
			items=(el**)realloc(items,maxSize*sizeof(el*));
			memset(&items[pms],0,(maxSize-pms)*sizeof(el*));
		}
		if(ndx-off<cnt && items[ndx-off])
			items[ndx-off]->downcount();
		if(i!=NULL)
			i->upcount();
		items[ndx-off]=i;
		if(ndx-off+1>cnt)
			cnt=ndx-off+1;

		return ndx;
	}

	//! Remove item from vector
	Bool remove(el *i)
	{
		base::ScopedLock lock(critSec);

		for(Size j=0;j<cnt;j++)
		{
			if(items[j]==i)
				return remove(j+off);
		}

		return false;
	}

	//! Remove item at specified index from vector
	Bool remove(Size at)
	{
		base::ScopedLock lock(critSec);

		if(at<off || at>=off+cnt)
			return false;

		at-=off;
		if(items[at]!=NULL)
		{
			el *it=items[at];
			items[at]=NULL;
			it->downcount();
		}
		Size j;
		for(j=0;items[j]==NULL && j<cnt;j++);
		if(j>0)
		{
			if(cnt>j)
				memmove(&items[0],&items[j],(cnt-j)*sizeof(el*));
			off+=j;
			cnt-=j;
		}

		return true;
	}

	//! Clear vector
	Bool clear()
	{
		base::ScopedLock lock(critSec);

		for(Size i=0;i<cnt;i++)
		{
			if(items[i])
			{
				el *it=items[i];
				items[i]=NULL;
				it->downcount();
			}
		}
		off=cnt;
		cnt=0;

		return true;
	}

	//! Get size of vector
	Size size() 
	{ 		
		base::ScopedLock lock(critSec);

		return cnt+off; 		
	}

	//! Get offset of first non-null item of vector
	Size offset() 
	{ 
		base::ScopedLock lock(critSec);

		return off; 
	}

	//! Get indexed item from vector
	el* operator[](Size i) 
	{ 
		base::ScopedLock lock(critSec);

		return (i>=off && i<off+cnt) ? items[i-off] : NULL; 
	}
};

//! Fixed-size array
/*! \ingroup internal
*/
template<typename el> class VarArray : public Object
{
	IDENTIFYNOREG(VarArray);
protected:
	//! Items in array
	Ptr<el> *items;
	//! Number of items in array
	Size cnt;
public:
	//! Constructor
	VarArray(Size s=0) 
	{ 
		if(s)
		{
			cnt=s; 
			items=new Ptr<el>[cnt]; 
		}
		else
		{
			cnt=0;
			items=NULL;
		}
	}
	//! Destructor
	~VarArray()
	{
		if(items)
			delete[] items;
	}
	//! Reset size of array
	/*! \param size New size of array
	*/
	void setSize(Size size)
	{
		if(cnt)
		{
			dump(0,"Cannot resize arrays");
			return;
		}
		cnt=size;
		items=new Ptr<el>[cnt]; 
	}

	//! Get indexed item from array
	Ptr<el>& operator[](Size i) { return items[i]; }

	//! Get size of array
	Size size() { return cnt; }
};

}

#endif
