/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*!	\file baseserial.h
	\brief Serialization classes

	These classes provide writers for binary files.
*/

#ifndef INCLUDED_BASE_SERIAL_H
#define INCLUDED_BASE_SERIAL_H

namespace base
{
	//! OutputStream with binary data format
	/*! \ingroup internal
	*/
	template<typename Writer> class BinaryOutputStream : public OutputStream
	{
		Writer *writer;
	public:
		BinaryOutputStream(Writer *w)
		{
			writer=w;
		}
		virtual void writeInt8(const Int8& item)
		{
			writer->writeBytes(&item,sizeof(Int8));
		}
		virtual void writeUInt8(const UInt8& item)
		{
			writer->writeBytes(&item,sizeof(UInt8));
		}
		virtual void writeInt16(const Int16& item)
		{
			writer->writeBytes(&item,sizeof(Int16));
		}
		virtual void writeUInt16(const UInt16& item)
		{
			writer->writeBytes(&item,sizeof(UInt16));
		}
		virtual void writeInt32(const Int32& item)
		{
			writer->writeBytes(&item,sizeof(Int32));
		}
		virtual void writeUInt32(const UInt32& item)
		{
			writer->writeBytes(&item,sizeof(UInt32));
		}
		virtual void writeInt64(const Int64& item)
		{
			writer->writeBytes(&item,sizeof(Int64));
		}
		virtual void writeUInt64(const UInt64& item)
		{
			writer->writeBytes(&item,sizeof(UInt64));
		}
		virtual void writeFloat(const Float& item)
		{
			writer->writeBytes(&item,sizeof(Float));
		}
		virtual void writeDouble(const Double& item)
		{
			writer->writeBytes(&item,sizeof(Double));
		}
		virtual void writeBool(const Bool& item)
		{
			writer->writeBytes(&item,1);
		}
		virtual void writeArray(const void *data, const Size count, const Size size)
		{
			writer->writeBytes(data,count*size);
		}
		virtual void writeString(const std::string& item)
		{
			XPSize s=(XPSize)item.length();
			typehandler::write(this,s);
			writer->writeBytes(item.c_str(),s);
		}

		virtual Bool write(const Object *item)
		{
			refs.addElement(item);
			for(Size i=1;i<refs.size();i++)
				refs.elementAt(i)->writeV(this);

			return writer->flush(refs);
		}
	};

	//! InputStream with binary data format
	/*! \ingroup internal
	*/
	template<typename Reader> class BinaryInputStream : public InputStream
	{
		Reader *reader;
	public:
		BinaryInputStream(Reader *r)
		{
			reader=r;
		}
		virtual void readInt8(Int8& item)
		{
			reader->readBytes(&item,sizeof(Int8));
		}
		virtual void readUInt8(UInt8& item)
		{
			reader->readBytes(&item,sizeof(UInt8));
		}
		virtual void readInt16(Int16& item)
		{
			reader->readBytes(&item,sizeof(Int16));
		}
		virtual void readUInt16(UInt16& item)
		{
			reader->readBytes(&item,sizeof(UInt16));
		}
		virtual void readInt32(Int32& item)
		{
			reader->readBytes(&item,sizeof(Int32));
		}
		virtual void readUInt32(UInt32& item)
		{
			reader->readBytes(&item,sizeof(UInt32));
		}
		virtual void readInt64(Int64& item)
		{
			reader->readBytes(&item,sizeof(Int64));
		}
		virtual void readUInt64(UInt64& item)
		{
			reader->readBytes(&item,sizeof(UInt64));
		}
		virtual void readFloat(Float& item)
		{
			reader->readBytes(&item,sizeof(Float));
		}
		virtual void readDouble(Double& item)
		{
			reader->readBytes(&item,sizeof(Double));
		}
		virtual void readBool(Bool& item)
		{
			reader->readBytes(&item,1);
		}
		virtual void readArray(void *data, const Size count, const Size size)
		{
			reader->readBytes(data,count*size);
		}
		virtual void readString(std::string& item)
		{
			XPSize s;
			typehandler::read(this,s);
			char *tmp=(char *)alloca(s+1);
			reader->readBytes(tmp,s);
			tmp[s]=0;
			item=tmp;
		}
		virtual Object *read()
		{
			if(!reader->init(refs))
				return NULL;

			for(Size i=1;i<refs.size();i++)
				refs.elementAt(i)->readV(this);

			return refs.elementAt(1);
		}
	};

	//! InputStream with binary data format, performs byte swapping
	/*! \ingroup internal
	*/
	template<typename Reader> class SwappingBinaryInputStream : public InputStream
	{
		Reader *reader;
	public:
		SwappingBinaryInputStream(Reader *r)
		{
			reader=r;
		}
		virtual void readInt8(Int8& item)
		{
			reader->readBytes(&item,sizeof(Int8));
		}
		virtual void readUInt8(UInt8& item)
		{
			reader->readBytes(&item,sizeof(UInt8));
		}
		virtual void readInt16(Int16& item)
		{
			reader->readBytes(&item,sizeof(Int16));
			byteswapper::swap(item);
		}
		virtual void readUInt16(UInt16& item)
		{
			reader->readBytes(&item,sizeof(UInt16));
			byteswapper::swap(item);
		}
		virtual void readInt32(Int32& item)
		{
			reader->readBytes(&item,sizeof(Int32));
			byteswapper::swap(item);
		}
		virtual void readUInt32(UInt32& item)
		{
			reader->readBytes(&item,sizeof(UInt32));
			byteswapper::swap(item);
		}
		virtual void readInt64(Int64& item)
		{
			reader->readBytes(&item,sizeof(Int64));
			byteswapper::swap(item);
		}
		virtual void readUInt64(UInt64& item)
		{
			reader->readBytes(&item,sizeof(UInt64));
			byteswapper::swap(item);
		}
		virtual void readFloat(Float& item)
		{
			reader->readBytes(&item,sizeof(Float));
			byteswapper::swap(item);
		}
		virtual void readDouble(Double& item)
		{
			reader->readBytes(&item,sizeof(Double));
			byteswapper::swap(item);
		}
		virtual void readBool(Bool& item)
		{
			reader->readBytes(&item,1);
		}
		//! Read an array of fixed size simple elements
		/*! This should actually be templated to the size of the array elements,
			in order to avoid the fatal error case when the size is something
			abnormal, but this is not an option since the method is virtual.
			NOTE: consider replacing this with 4 functions, for size=1,2,4 and 8.
		*/
		virtual void readArray(void *data, const Size count, const Size size)
		{
			reader->readBytes(data,count*size);
			if(size==1)
			{
			}
			else if(size==2)
			{
				Int16 *d=(Int16*)data;
				for(Size i=0;i<count;i++)
					byteswapper::swapp<2>(d++);
			}
			else if(size==4)
			{
				Int32 *d=(Int32*)data;
				for(Size i=0;i<count;i++)
					byteswapper::swapp<4>(d++);
			}
			else if(size==8)
			{
				Int64 *d=(Int64*)data;
				for(Size i=0;i<count;i++)
					byteswapper::swapp<8>(d++);
			}
			else
				fatalError("Invalid data size on SwappingBinaryInputStream::readArray: %d",size);
		}
		virtual void readString(std::string& item)
		{
			XPSize s;
			typehandler::read(this,s);
			char *tmp=(char *)alloca(s+1);
			reader->readBytes(tmp,s);
			tmp[s]=0;
			item=tmp;
		}
		virtual Object *read()
		{
			if(!reader->init(refs))
				return NULL;

			for(Size i=1;i<refs.size();i++)
				refs.elementAt(i)->readV(this);

			return refs.elementAt(1);
		}
	};
}

#endif

